#!/usr/bin/env python
# -*-coding: utf-8 -*-
# vim: sw=4 ts=4 expandtab ai
#
#    Copyright 2022 VK Cloud.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import print_function

import logging
import netaddr
import time

from google.protobuf.any_pb2 import Any
import grpc

from evpn_connector.bgp.generated import attribute_pb2
from evpn_connector.bgp.generated import gobgp_pb2
from evpn_connector.bgp.generated import gobgp_pb2_grpc
from evpn_connector.common import constants
from evpn_connector.service import objects as evpnobj

LOG = logging.getLogger(__name__)
BGP_ORIGIN = 2
EXT_COM_RT_SUBTYPE = 2
VXLAN_ENCAP_TYPE = 8
PMSI_TYPE_INGRESS_REPLICATION = 6
TYPE_5_GW_ADDRESS = "0.0.0.0"
PROTOBUF_NIL = "<nil>"
MPREACHNLRI_TYPE_NAME = "type.googleapis.com/apipb.MpReachNLRIAttribute"
EXTC_TYPE_NAME = "type.googleapis.com/apipb.ExtendedCommunitiesAttribute"
TOSE_TYPE_NAME = "type.googleapis.com/apipb.TwoOctetAsSpecificExtended"
PMSI_TYPE_NAME = "type.googleapis.com/apipb.PmsiTunnelAttribute"
ROUTER_MAC_EXT_TYPE_NAME = "type.googleapis.com/apipb.RouterMacExtended"

STUB = None

"""
# install golang>1.11, protoc (brew install protobuf), protoc-gen-go

$ pip install grpcio-tools
$ git clone git://github.com/osrg/gobgp
$ cd gobgp

change api/*.proto:
```
diff --git a/api/attribute.proto b/api/attribute.proto
index 4cf844f..7bf9452 100644
--- a/api/attribute.proto
+++ b/api/attribute.proto
@@ -21,7 +21,7 @@

 syntax = "proto3";

-import "any/any.proto";
+import "google/protobuf/any.proto";
 import "gobgp.proto";

 package gobgpapi;
diff --git a/api/gobgp.proto b/api/gobgp.proto
index b61b475..52dd9da 100644
--- a/api/gobgp.proto
+++ b/api/gobgp.proto
@@ -21,9 +21,9 @@

 syntax = "proto3";

-import "any/any.proto";
-import "empty/empty.proto";
-import "timestamp/timestamp.proto";
+import "google/protobuf/any.proto";
+import "google/protobuf/empty.proto";
+import "google/protobuf/timestamp.proto";

 package gobgpapi;


```

$ mkdir ./py_out
$ python -m grpc_tools.protoc -I./api --python_out=./py_out/ \
--grpc_python_out=./py_out/ \
api/gobgp.proto api/attribute.proto api/capability.proto

Voila, you've got everything you need in ./py_out !

grpc_tools generate absolute imports of generated files related to root,
so maybe you need to replace them with
tools/replace-protobuf-local-imports.sh path/to/generated/files
"""


def unpack_protobuf_message(struct, module=attribute_pb2):
    """Unpacks message with `Any` type.

    After that you can use returned structure like common python's
    object. Searches type of needed message in specified module than
    unpacks bytes of specified structure to this defined message.

    Args:
        struct(google.protobuf.any_pb2.Any): structure which we want
            to unpack
        module: generated python's module with definitions of protobuf
            messages

    Returns:
        gobgp_pb2.<SomeClassOfMessage>

    """
    if type(struct) != Any:
        raise TypeError("You should specify struct with type Any to unpack")

    message = getattr(module, struct.type_url.split(".")[-1])()
    struct.Unpack(message)

    return message


def _simplify_unicast_struct(struct):
    res = {}
    # Need for extract: mac, rd_admin, ip, next_hop, vni, rt_asn, rt_num
    # Convert protobuf message to dict

    nlri = unpack_protobuf_message(struct.paths[0].nlri)

    rd = unpack_protobuf_message(nlri.rd)

    res["mac"] = nlri.mac_address
    # Extract rd_admin only if it's AS_number. Needed only for local announces
    try:
        res["rd_admin"] = int(rd.admin)
    except ValueError:
        LOG.debug(
            "Get macadv announce with mac: %s and non-ASN RD: %s",
            nlri.mac_address,
            rd.admin,
        )
    res["ip"] = nlri.ip_address
    if res["ip"] == PROTOBUF_NIL:
        res["ip"] = ""
    # Get only first label from announce
    res["vni"] = nlri.labels[0]

    res["rt"] = []
    # All remote clients has type from EncapExtended attribute
    # Curently we support only VXLAN
    res["type"] = constants.EVPN_EDGE_TYPE_VXLAN

    pattrs = struct.paths[0].pattrs
    # Pattrs may be list with random order
    for pattr in pattrs:
        if pattr.type_url == MPREACHNLRI_TYPE_NAME:
            # TODO(aleks.popov): Add ecmp support
            pa = unpack_protobuf_message(pattr)

            res["next_hop"] = pa.next_hops[0]

        elif pattr.type_url == EXTC_TYPE_NAME:
            pa = unpack_protobuf_message(pattr)
            for community in pa.communities:
                if community.type_url == TOSE_TYPE_NAME:
                    comm = unpack_protobuf_message(community)

                    res["rt"].append((comm.asn, comm.local_admin))
    return evpnobj.ClientEdge.from_dict(res)


def _simplify_prefix_struct(struct):
    result = []
    for path in struct.paths:
        # Need for extract: prefix, rd_admin, next_hop, vni, rt_asn, rt_num

        nlri = unpack_protobuf_message(path.nlri)

        rd = unpack_protobuf_message(nlri.rd)

        res = {
            # For older protobuf version names in nlri may be ipPrefix and
            # ipPrefixLen
            "prefix": nlri.ip_prefix,
            "prefix_len": nlri.ip_prefix_len,
            "vni": nlri.label,
            "rt": [],
            # All remote clients has type from EncapExtended attribute Curently
            # we support only VXLAN
            "type": constants.EVPN_EDGE_TYPE_VXLAN,
        }

        # Extract rd_admin only if it's AS_number. Needed only for local
        # announces
        try:
            res["rd_admin"] = int(rd.admin)
        except ValueError:
            LOG.debug(
                "Get Prefix announce with ip/mask: %s/%s and non-ASN RD: %s",
                # FIXME: For older protobuf version names in nlri may be
                # ipPrefix and ipPrefixLen
                nlri.ip_prefix,
                nlri.ip_prefix_len,
                rd.admin,
            )

        pattrs = path.pattrs
        # Pattrs may be list with random order
        for pattr in pattrs:
            if pattr.type_url == MPREACHNLRI_TYPE_NAME:
                pa = unpack_protobuf_message(pattr)
                res["next_hop"] = pa.next_hops[0]
            elif pattr.type_url == EXTC_TYPE_NAME:
                pa = unpack_protobuf_message(pattr)
                for community in pa.communities:
                    if community.type_url == TOSE_TYPE_NAME:
                        comm = unpack_protobuf_message(community)
                        res["rt"].append((comm.asn, comm.local_admin))
                    if community.type_url == ROUTER_MAC_EXT_TYPE_NAME:
                        comm = unpack_protobuf_message(community)
                        res["router_mac"] = comm.mac

        if "router_mac" not in res:
            LOG.warning("Not found RouterMacExtended in announce. Use default")
            res["router_mac"] = constants.TYPE_5_DEFAULT_ROUTER_MAC_EXTENDED

        result.append(evpnobj.ClientEdgePrefix.from_dict(res))

    return result


def _simplify_multicast_struct(multicast_struct):
    # Convert protobuf message to dict

    res = {}
    # Need for extract: rd_admin, next_hop, vni, rt_asn, rt_num
    nlri = unpack_protobuf_message(multicast_struct.paths[0].nlri)
    rd = unpack_protobuf_message(nlri.rd)

    # Extract rd_admin only if it's AS_number. Needed only for local announces
    try:
        res["rd_admin"] = int(rd.admin)
    except ValueError:
        LOG.debug("Get multicast announce with non-ASN RD: %s", rd.admin)
    res["rt"] = []

    pattrs = multicast_struct.paths[0].pattrs
    # Pattrs may be list with random order
    for pattr in pattrs:
        if pattr.type_url == MPREACHNLRI_TYPE_NAME:
            # TODO(aleks.popov): Add ecmp support
            pa = unpack_protobuf_message(pattr)
            res["next_hop"] = pa.next_hops[0]
        elif pattr.type_url == EXTC_TYPE_NAME:
            pa = unpack_protobuf_message(pattr)
            for community in pa.communities:
                if community.type_url == TOSE_TYPE_NAME:
                    comm = unpack_protobuf_message(community)

                    res["rt"].append((comm.asn, comm.local_admin))
        elif pattr.type_url == PMSI_TYPE_NAME:
            pa = unpack_protobuf_message(pattr)
            res["vni"] = pa.label
    return evpnobj.VirtNet.from_dict(res)


class BGPClient(object):
    family = gobgp_pb2.Family(
        afi=gobgp_pb2.Family.AFI_L2VPN, safi=gobgp_pb2.Family.SAFI_EVPN
    )
    simplify_map = {
        constants.EVPN_MACADV_TYPE: _simplify_unicast_struct,
        constants.EVPN_MULTICAST_TYPE: _simplify_multicast_struct,
        constants.EVPN_PREFIX_TYPE: _simplify_prefix_struct,
    }

    def __init__(self, gobgp_channel, grpc_timeout_sec, router_mac_type5):
        self._router_mac_type5 = router_mac_type5
        self.gobgp_channel = gobgp_channel
        self.timeout_seconds = grpc_timeout_sec
        self.bgp_origin = Any()
        self.bgp_origin.Pack(attribute_pb2.OriginAttribute(origin=BGP_ORIGIN))

        global STUB
        if not STUB:
            # channel and stub are thread-safe
            channel = grpc.insecure_channel(self.gobgp_channel)
            STUB = gobgp_pb2_grpc.GobgpApiStub(channel)
        self.stub = STUB

    @staticmethod
    def _get_ext_comm(rt, tunnel_type=VXLAN_ENCAP_TYPE, router_mac=""):
        # BGP Extended communities
        # Route target
        tgts = []
        for asn, num in rt.targets:
            rt_dict = {
                "sub_type": EXT_COM_RT_SUBTYPE,
                "asn": asn,
                "local_admin": num,
                "is_transitive": True,
            }
            rt = Any()
            rt.Pack(attribute_pb2.TwoOctetAsSpecificExtended(**rt_dict))
            tgts.append(rt)
        # Encapsulation type
        encap = Any()
        encap.Pack(attribute_pb2.EncapExtended(tunnel_type=tunnel_type))

        rtr_mac = Any()
        if router_mac:
            rtr_mac.Pack(attribute_pb2.RouterMacExtended(mac=router_mac))

        ext_community = Any()
        if router_mac:
            ext_community.Pack(
                attribute_pb2.ExtendedCommunitiesAttribute(
                    communities=tgts + [encap] + [rtr_mac]
                )
            )
        else:
            ext_community.Pack(
                attribute_pb2.ExtendedCommunitiesAttribute(
                    communities=tgts + [encap]
                )
            )
        return ext_community

    @staticmethod
    def _get_rd(rd_asn, rd_num):
        rd_addr = Any()
        rd_addr.Pack(
            attribute_pb2.RouteDistinguisherTwoOctetASN(
                admin=rd_asn,
                assigned=rd_num,
            )
        )
        return rd_addr

    def prepare_unicast_path(self, edge):
        """Prepare evpn Type 2 route"""
        # Route information
        rd = self._get_rd(edge.as_number, edge.rd_num)
        nlri = Any()
        nlri.Pack(
            attribute_pb2.EVPNMACIPAdvertisementRoute(
                mac_address=edge.mac,
                ip_address=edge.ip,
                labels=[edge.vni],
                rd=rd,
                esi=attribute_pb2.EthernetSegmentIdentifier(),
            )
        )

        # Route params
        # BGP Extended communities
        ext_community = self._get_ext_comm(edge.rt)

        # NLRI
        mp_reach_nlri = Any()
        mp_reach_nlri.Pack(
            attribute_pb2.MpReachNLRIAttribute(
                family=self.family,
                next_hops=[edge.next_hop],
                nlris=[nlri],
            )
        )

        path = gobgp_pb2.Path(
            nlri=nlri,
            pattrs=[self.bgp_origin, ext_community, mp_reach_nlri],
            family=self.family,
        )
        return path

    def add_unicast_path(self, edge):
        path = self.prepare_unicast_path(edge)
        LOG.info("Add macadv announce: %s", edge)
        return self.stub.AddPath(
            gobgp_pb2.AddPathRequest(
                table_type=gobgp_pb2.GLOBAL,
                path=path,
            ),
            self.timeout_seconds,
        )

    def del_unicast_path(self, edge):
        path = self.prepare_unicast_path(edge)
        LOG.info("Del macadv announce: %s", edge)
        return self.stub.DeletePath(
            gobgp_pb2.DeletePathRequest(
                table_type=gobgp_pb2.GLOBAL,
                family=self.family,
                path=path,
            ),
            self.timeout_seconds,
        )

    def prepare_multicast_path(self, vnet):
        """Add evpn Type 3 route"""
        # Route information
        rd = self._get_rd(vnet.as_number, vnet.rd_num)
        nlri = Any()
        nlri.Pack(
            attribute_pb2.EVPNInclusiveMulticastEthernetTagRoute(
                ip_address=vnet.next_hop,
                rd=rd,
            )
        )

        pmsi_dict = {
            "type": PMSI_TYPE_INGRESS_REPLICATION,
            "label": vnet.vni,
            "id": netaddr.IPAddress(vnet.next_hop).packed,
        }
        pmsi = Any()
        pmsi.Pack(attribute_pb2.PmsiTunnelAttribute(**pmsi_dict))

        # Route params
        ext_community = self._get_ext_comm(vnet.rt)

        # NLRI
        mp_reach_nlri = Any()
        mp_reach_nlri.Pack(
            attribute_pb2.MpReachNLRIAttribute(
                family=self.family,
                next_hops=[vnet.next_hop],
                nlris=[nlri],
            )
        )

        path = gobgp_pb2.Path(
            nlri=nlri,
            pattrs=[self.bgp_origin, pmsi, ext_community, mp_reach_nlri],
            family=self.family,
        )
        return path

    def add_multicast_path(self, vnet):
        LOG.info("Add multicast announce: %s", vnet)
        path = self.prepare_multicast_path(vnet)
        return self.stub.AddPath(
            gobgp_pb2.AddPathRequest(
                table_type=gobgp_pb2.GLOBAL,
                path=path,
            ),
            self.timeout_seconds,
        )

    def del_multicast_path(self, vnet):
        LOG.info("Del multicast announce: %s", vnet)
        path = self.prepare_multicast_path(vnet)
        return self.stub.DeletePath(
            gobgp_pb2.DeletePathRequest(
                table_type=gobgp_pb2.GLOBAL,
                family=self.family,
                path=path,
            ),
            self.timeout_seconds,
        )

    def prepare_prefix_path(self, edge_prefix):
        """Prepare evpn Type 5 route"""
        # Route information
        rd = self._get_rd(edge_prefix.as_number, edge_prefix.rd_num)
        nlri = Any()
        nlri.Pack(
            attribute_pb2.EVPNIPPrefixRoute(
                # FIXME: For older protobuf version names in nlri may be
                # ipPrefix and ipPrefixLen and gwAddress
                ip_prefix=edge_prefix.prefix,
                ip_prefix_len=edge_prefix.prefix_len,
                gw_address=TYPE_5_GW_ADDRESS,
                label=edge_prefix.vni,
                rd=rd,
                esi=attribute_pb2.EthernetSegmentIdentifier(),
            )
        )

        # Route params
        # BGP Extended communities for Type 5 (with RouterMacExtended)
        ext_community = self._get_ext_comm(
            rt=edge_prefix.rt, router_mac=self._router_mac_type5
        )

        # NLRI
        mp_reach_nlri = Any()
        mp_reach_nlri.Pack(
            attribute_pb2.MpReachNLRIAttribute(
                family=self.family,
                next_hops=[edge_prefix.next_hop],
                nlris=[nlri],
            )
        )

        path = gobgp_pb2.Path(
            nlri=nlri,
            pattrs=[self.bgp_origin, ext_community, mp_reach_nlri],
            family=self.family,
        )

        return path

    def add_prefix_path(self, edge_prefix):
        path = self.prepare_prefix_path(edge_prefix)
        LOG.info("Add Prefix announce: %s", edge_prefix)
        return self.stub.AddPath(
            gobgp_pb2.AddPathRequest(
                table_type=gobgp_pb2.GLOBAL,
                path=path,
            ),
            self.timeout_seconds,
        )

    def del_prefix_path(self, edge_prefix):
        path = self.prepare_prefix_path(edge_prefix)
        LOG.info("Del Prefix announce: %s", edge_prefix)
        return self.stub.DeletePath(
            gobgp_pb2.DeletePathRequest(
                table_type=gobgp_pb2.GLOBAL,
                family=self.family,
                path=path,
            ),
            self.timeout_seconds,
        )

    @staticmethod
    def filter_local_announces(objs, local_next_hop):
        client_edges, virt_nets, client_prefixes = objs

        local_edges = set()
        remote_edges = set()
        for edge in client_edges:
            if edge.next_hop == local_next_hop:
                local_edges.add(edge)
            else:
                remote_edges.add(edge)

        local_virt_nets = set()
        remote_virt_nets = set()
        for vnet in virt_nets:
            if vnet.next_hop == local_next_hop:
                local_virt_nets.add(vnet)
            else:
                remote_virt_nets.add(vnet)

        local_client_prefixes = set()
        remote_client_prefixes = set()
        for cprefix in client_prefixes:
            if cprefix.next_hop == local_next_hop:
                local_client_prefixes.add(cprefix)
            else:
                remote_client_prefixes.add(cprefix)

        return (
            local_edges,
            local_virt_nets,
            remote_edges,
            remote_virt_nets,
            local_client_prefixes,
            remote_client_prefixes,
        )

    def get_paths(self, local_next_hop=None):
        return self.filter_local_announces(self._get_paths(), local_next_hop)

    def _get_paths_from_gobgp(self, prefix_type):
        start_time = time.time()
        req = self.stub.ListPath(
            gobgp_pb2.ListPathRequest(
                table_type=gobgp_pb2.GLOBAL,
                family=self.family,
                prefixes=[gobgp_pb2.TableLookupPrefix(prefix=prefix_type)],
            ),
            self.timeout_seconds,
        )
        LOG.debug(
            "Get data from gobgp for %s type: %0.6f sec",
            prefix_type,
            time.time() - start_time,
        )
        return req

    def _get_paths(self):
        """Get paths from GoBGP"""
        return (
            self._get_obj(constants.EVPN_MACADV_TYPE),
            self._get_obj(constants.EVPN_MULTICAST_TYPE),
            self._get_obj(constants.EVPN_PREFIX_TYPE),
        )

    def _get_obj(self, prefix_type):
        """Get paths from GoBGP"""
        res = []
        req = self._get_paths_from_gobgp(prefix_type=prefix_type)
        start_time = time.time()
        for line in req:
            simplify_func = self.simplify_map[prefix_type]
            func_res = simplify_func(line.destination)
            if type(func_res) is list:
                res += func_res
            else:
                res.append(func_res)
        LOG.debug(
            "Parse %d prefixes from gobgp for %s type: %0.6f sec",
            len(res),
            prefix_type,
            time.time() - start_time,
        )
        return res

    def _extract_response_from_grpc(self, grpc_channel):
        response = [item for item in grpc_channel]
        if len(response) > 0:
            return response
        return None

    def add_defined_set(self, defined_set):
        self.stub.AddDefinedSet(
            gobgp_pb2.AddDefinedSetRequest(
                defined_set=defined_set.to_grpc(),
            ),
            self.timeout_seconds,
        )
        LOG.debug("Defined set %s added to GoBGP", defined_set.name)

    def delete_defined_set(self, defined_set, full_deletion=False):
        self.stub.DeleteDefinedSet(
            gobgp_pb2.DeleteDefinedSetRequest(
                defined_set=defined_set.to_grpc(),
                all=full_deletion,
            ),
            self.timeout_seconds,
        )
        LOG.debug("Defined set %s removed from GoBGP", defined_set.name)

    def get_defined_set(self, defined_type, defined_set_name):
        defined_set_channel = self.stub.ListDefinedSet(
            gobgp_pb2.ListDefinedSetRequest(
                defined_type=defined_type,
                name=defined_set_name,
            ),
            self.timeout_seconds,
        )
        resp = self._extract_response_from_grpc(defined_set_channel)
        if resp is None:
            return None
        return evpnobj.DefinedSet.from_grpc(resp[0].defined_set)

    def add_statement(self, statement):
        self.stub.AddStatement(
            gobgp_pb2.AddStatementRequest(
                statement=statement.to_grpc(),
            ),
            self.timeout_seconds,
        )
        LOG.debug("Statement %s added to GoBGP", statement.name)

    def delete_statement(self, statement, full_deletion=False):
        self.stub.DeleteStatement(
            gobgp_pb2.DeleteStatementRequest(
                statement=statement.to_grpc(),
                all=full_deletion,
            ),
            self.timeout_seconds,
        )
        LOG.debug("Statement %s removed from GoBGP", statement.name)

    def get_statement(self, statement_name):
        statement_channel = self.stub.ListStatement(
            gobgp_pb2.ListStatementRequest(name=statement_name),
            self.timeout_seconds,
        )
        resp = self._extract_response_from_grpc(statement_channel)
        if resp is None:
            return None
        return evpnobj.Statement.from_grpc(resp[0].statement)

    def add_policy(self, policy):
        self.stub.AddPolicy(
            gobgp_pb2.AddPolicyRequest(
                policy=policy.to_grpc(),
                refer_existing_statements=True,
            ),
            self.timeout_seconds,
        )
        LOG.debug("Policy %s added to GoBGP", policy.name)

    def delete_policy(self, policy, full_deletion=False):
        self.stub.DeletePolicy(
            gobgp_pb2.DeletePolicyRequest(
                policy=policy.to_grpc(),
                preserve_statements=True,
                all=full_deletion,
            ),
            self.timeout_seconds,
        )
        LOG.debug("Policy %s removed from GoBGP", policy.name)

    def get_policy(self, policy_name):
        policy_channel = self.stub.ListPolicy(
            gobgp_pb2.ListPolicyRequest(
                name=policy_name,
            ),
            self.timeout_seconds,
        )
        resp = self._extract_response_from_grpc(policy_channel)
        if resp is None:
            return None
        return evpnobj.Policy.from_grpc(resp[0].policy)

    def add_policy_assignment(self, policy_assignment):
        self.stub.AddPolicyAssignment(
            gobgp_pb2.AddPolicyAssignmentRequest(
                assignment=policy_assignment.to_grpc(),
            ),
            self.timeout_seconds,
        )
        LOG.debug("Policy assignment %s add to GoBGP", policy_assignment.name)

    def delete_policy_assignment(self, policy_assignment, full_deletion=False):
        self.stub.DeletePolicyAssignment(
            gobgp_pb2.DeletePolicyAssignmentRequest(
                assignment=policy_assignment.to_grpc(),
                all=full_deletion,
            ),
            self.timeout_seconds,
        )
        LOG.debug(
            "Policy assignment %s removed from GoBGP", policy_assignment.name
        )

    def get_policy_assignment(self, pa_name, pa_direction):
        policy_assignment_channel = self.stub.ListPolicyAssignment(
            gobgp_pb2.ListPolicyAssignmentRequest(
                name=pa_name,
                direction=pa_direction,
            ),
            self.timeout_seconds,
        )
        resp = self._extract_response_from_grpc(policy_assignment_channel)
        if resp is None or len(resp[0].assignment.policies) == 0:
            return None
        return evpnobj.PolicyAssignment.from_grpc(resp[0].assignment)

    def list_peer(self, peer_address):
        list_peers_channel = self.stub.ListPeer(
            gobgp_pb2.ListPeerRequest(
                address=peer_address,
            ),
            self.timeout_seconds,
        )
        resp = self._extract_response_from_grpc(list_peers_channel)
        return resp

    def reset_peer(self, address, soft, direction):
        self.stub.ResetPeer(
            gobgp_pb2.ResetPeerRequest(
                address=address,
                soft=soft,
                direction=direction,
            ),
            self.timeout_seconds,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: capability.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import evpn_connector.bgp.generated.gobgp_pb2 as gobgp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x63\x61pability.proto\x12\x05\x61pipb\x1a\x0bgobgp.proto\"8\n\x17MultiProtocolCapability\x12\x1d\n\x06\x66\x61mily\x18\x01 \x01(\x0b\x32\r.apipb.Family\"\x18\n\x16RouteRefreshCapability\"\x1d\n\x1b\x43\x61rryingLabelInfoCapability\"k\n\x1e\x45xtendedNexthopCapabilityTuple\x12\"\n\x0bnlri_family\x18\x01 \x01(\x0b\x32\r.apipb.Family\x12%\n\x0enexthop_family\x18\x02 \x01(\x0b\x32\r.apipb.Family\"R\n\x19\x45xtendedNexthopCapability\x12\x35\n\x06tuples\x18\x01 \x03(\x0b\x32%.apipb.ExtendedNexthopCapabilityTuple\"N\n\x1eGracefulRestartCapabilityTuple\x12\x1d\n\x06\x66\x61mily\x18\x01 \x01(\x0b\x32\r.apipb.Family\x12\r\n\x05\x66lags\x18\x02 \x01(\r\"o\n\x19GracefulRestartCapability\x12\r\n\x05\x66lags\x18\x01 \x01(\r\x12\x0c\n\x04time\x18\x02 \x01(\r\x12\x35\n\x06tuples\x18\x03 \x03(\x0b\x32%.apipb.GracefulRestartCapabilityTuple\"%\n\x16\x46ourOctetASNCapability\x12\x0b\n\x03\x61sn\x18\x01 \x01(\r\"\x9c\x01\n\x16\x41\x64\x64PathCapabilityTuple\x12\x1d\n\x06\x66\x61mily\x18\x01 \x01(\x0b\x32\r.apipb.Family\x12\x30\n\x04mode\x18\x02 \x01(\x0e\x32\".apipb.AddPathCapabilityTuple.Mode\"1\n\x04Mode\x12\x08\n\x04NONE\x10\x00\x12\x0b\n\x07RECEIVE\x10\x01\x12\x08\n\x04SEND\x10\x02\x12\x08\n\x04\x42OTH\x10\x03\"B\n\x11\x41\x64\x64PathCapability\x12-\n\x06tuples\x18\x01 \x03(\x0b\x32\x1d.apipb.AddPathCapabilityTuple\" \n\x1e\x45nhancedRouteRefreshCapability\"e\n\'LongLivedGracefulRestartCapabilityTuple\x12\x1d\n\x06\x66\x61mily\x18\x01 \x01(\x0b\x32\r.apipb.Family\x12\r\n\x05\x66lags\x18\x02 \x01(\r\x12\x0c\n\x04time\x18\x03 \x01(\r\"d\n\"LongLivedGracefulRestartCapability\x12>\n\x06tuples\x18\x01 \x03(\x0b\x32..apipb.LongLivedGracefulRestartCapabilityTuple\"\x1d\n\x1bRouteRefreshCiscoCapability\"8\n\x0e\x46qdnCapability\x12\x11\n\thost_name\x18\x01 \x01(\t\x12\x13\n\x0b\x64omain_name\x18\x02 \x01(\t\"5\n\x19SoftwareVersionCapability\x12\x18\n\x10software_version\x18\x01 \x01(\t\"0\n\x11UnknownCapability\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x0c\x42$Z\"github.com/osrg/gobgp/v3/api;apipbb\x06proto3')



_MULTIPROTOCOLCAPABILITY = DESCRIPTOR.message_types_by_name['MultiProtocolCapability']
_ROUTEREFRESHCAPABILITY = DESCRIPTOR.message_types_by_name['RouteRefreshCapability']
_CARRYINGLABELINFOCAPABILITY = DESCRIPTOR.message_types_by_name['CarryingLabelInfoCapability']
_EXTENDEDNEXTHOPCAPABILITYTUPLE = DESCRIPTOR.message_types_by_name['ExtendedNexthopCapabilityTuple']
_EXTENDEDNEXTHOPCAPABILITY = DESCRIPTOR.message_types_by_name['ExtendedNexthopCapability']
_GRACEFULRESTARTCAPABILITYTUPLE = DESCRIPTOR.message_types_by_name['GracefulRestartCapabilityTuple']
_GRACEFULRESTARTCAPABILITY = DESCRIPTOR.message_types_by_name['GracefulRestartCapability']
_FOUROCTETASNCAPABILITY = DESCRIPTOR.message_types_by_name['FourOctetASNCapability']
_ADDPATHCAPABILITYTUPLE = DESCRIPTOR.message_types_by_name['AddPathCapabilityTuple']
_ADDPATHCAPABILITY = DESCRIPTOR.message_types_by_name['AddPathCapability']
_ENHANCEDROUTEREFRESHCAPABILITY = DESCRIPTOR.message_types_by_name['EnhancedRouteRefreshCapability']
_LONGLIVEDGRACEFULRESTARTCAPABILITYTUPLE = DESCRIPTOR.message_types_by_name['LongLivedGracefulRestartCapabilityTuple']
_LONGLIVEDGRACEFULRESTARTCAPABILITY = DESCRIPTOR.message_types_by_name['LongLivedGracefulRestartCapability']
_ROUTEREFRESHCISCOCAPABILITY = DESCRIPTOR.message_types_by_name['RouteRefreshCiscoCapability']
_FQDNCAPABILITY = DESCRIPTOR.message_types_by_name['FqdnCapability']
_SOFTWAREVERSIONCAPABILITY = DESCRIPTOR.message_types_by_name['SoftwareVersionCapability']
_UNKNOWNCAPABILITY = DESCRIPTOR.message_types_by_name['UnknownCapability']
_ADDPATHCAPABILITYTUPLE_MODE = _ADDPATHCAPABILITYTUPLE.enum_types_by_name['Mode']
MultiProtocolCapability = _reflection.GeneratedProtocolMessageType('MultiProtocolCapability', (_message.Message,), {
  'DESCRIPTOR' : _MULTIPROTOCOLCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.MultiProtocolCapability)
  })
_sym_db.RegisterMessage(MultiProtocolCapability)

RouteRefreshCapability = _reflection.GeneratedProtocolMessageType('RouteRefreshCapability', (_message.Message,), {
  'DESCRIPTOR' : _ROUTEREFRESHCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.RouteRefreshCapability)
  })
_sym_db.RegisterMessage(RouteRefreshCapability)

CarryingLabelInfoCapability = _reflection.GeneratedProtocolMessageType('CarryingLabelInfoCapability', (_message.Message,), {
  'DESCRIPTOR' : _CARRYINGLABELINFOCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.CarryingLabelInfoCapability)
  })
_sym_db.RegisterMessage(CarryingLabelInfoCapability)

ExtendedNexthopCapabilityTuple = _reflection.GeneratedProtocolMessageType('ExtendedNexthopCapabilityTuple', (_message.Message,), {
  'DESCRIPTOR' : _EXTENDEDNEXTHOPCAPABILITYTUPLE,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.ExtendedNexthopCapabilityTuple)
  })
_sym_db.RegisterMessage(ExtendedNexthopCapabilityTuple)

ExtendedNexthopCapability = _reflection.GeneratedProtocolMessageType('ExtendedNexthopCapability', (_message.Message,), {
  'DESCRIPTOR' : _EXTENDEDNEXTHOPCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.ExtendedNexthopCapability)
  })
_sym_db.RegisterMessage(ExtendedNexthopCapability)

GracefulRestartCapabilityTuple = _reflection.GeneratedProtocolMessageType('GracefulRestartCapabilityTuple', (_message.Message,), {
  'DESCRIPTOR' : _GRACEFULRESTARTCAPABILITYTUPLE,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.GracefulRestartCapabilityTuple)
  })
_sym_db.RegisterMessage(GracefulRestartCapabilityTuple)

GracefulRestartCapability = _reflection.GeneratedProtocolMessageType('GracefulRestartCapability', (_message.Message,), {
  'DESCRIPTOR' : _GRACEFULRESTARTCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.GracefulRestartCapability)
  })
_sym_db.RegisterMessage(GracefulRestartCapability)

FourOctetASNCapability = _reflection.GeneratedProtocolMessageType('FourOctetASNCapability', (_message.Message,), {
  'DESCRIPTOR' : _FOUROCTETASNCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.FourOctetASNCapability)
  })
_sym_db.RegisterMessage(FourOctetASNCapability)

AddPathCapabilityTuple = _reflection.GeneratedProtocolMessageType('AddPathCapabilityTuple', (_message.Message,), {
  'DESCRIPTOR' : _ADDPATHCAPABILITYTUPLE,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.AddPathCapabilityTuple)
  })
_sym_db.RegisterMessage(AddPathCapabilityTuple)

AddPathCapability = _reflection.GeneratedProtocolMessageType('AddPathCapability', (_message.Message,), {
  'DESCRIPTOR' : _ADDPATHCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.AddPathCapability)
  })
_sym_db.RegisterMessage(AddPathCapability)

EnhancedRouteRefreshCapability = _reflection.GeneratedProtocolMessageType('EnhancedRouteRefreshCapability', (_message.Message,), {
  'DESCRIPTOR' : _ENHANCEDROUTEREFRESHCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.EnhancedRouteRefreshCapability)
  })
_sym_db.RegisterMessage(EnhancedRouteRefreshCapability)

LongLivedGracefulRestartCapabilityTuple = _reflection.GeneratedProtocolMessageType('LongLivedGracefulRestartCapabilityTuple', (_message.Message,), {
  'DESCRIPTOR' : _LONGLIVEDGRACEFULRESTARTCAPABILITYTUPLE,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.LongLivedGracefulRestartCapabilityTuple)
  })
_sym_db.RegisterMessage(LongLivedGracefulRestartCapabilityTuple)

LongLivedGracefulRestartCapability = _reflection.GeneratedProtocolMessageType('LongLivedGracefulRestartCapability', (_message.Message,), {
  'DESCRIPTOR' : _LONGLIVEDGRACEFULRESTARTCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.LongLivedGracefulRestartCapability)
  })
_sym_db.RegisterMessage(LongLivedGracefulRestartCapability)

RouteRefreshCiscoCapability = _reflection.GeneratedProtocolMessageType('RouteRefreshCiscoCapability', (_message.Message,), {
  'DESCRIPTOR' : _ROUTEREFRESHCISCOCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.RouteRefreshCiscoCapability)
  })
_sym_db.RegisterMessage(RouteRefreshCiscoCapability)

FqdnCapability = _reflection.GeneratedProtocolMessageType('FqdnCapability', (_message.Message,), {
  'DESCRIPTOR' : _FQDNCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.FqdnCapability)
  })
_sym_db.RegisterMessage(FqdnCapability)

SoftwareVersionCapability = _reflection.GeneratedProtocolMessageType('SoftwareVersionCapability', (_message.Message,), {
  'DESCRIPTOR' : _SOFTWAREVERSIONCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.SoftwareVersionCapability)
  })
_sym_db.RegisterMessage(SoftwareVersionCapability)

UnknownCapability = _reflection.GeneratedProtocolMessageType('UnknownCapability', (_message.Message,), {
  'DESCRIPTOR' : _UNKNOWNCAPABILITY,
  '__module__' : 'capability_pb2'
  # @@protoc_insertion_point(class_scope:apipb.UnknownCapability)
  })
_sym_db.RegisterMessage(UnknownCapability)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\"github.com/osrg/gobgp/v3/api;apipb'
  _MULTIPROTOCOLCAPABILITY._serialized_start=40
  _MULTIPROTOCOLCAPABILITY._serialized_end=96
  _ROUTEREFRESHCAPABILITY._serialized_start=98
  _ROUTEREFRESHCAPABILITY._serialized_end=122
  _CARRYINGLABELINFOCAPABILITY._serialized_start=124
  _CARRYINGLABELINFOCAPABILITY._serialized_end=153
  _EXTENDEDNEXTHOPCAPABILITYTUPLE._serialized_start=155
  _EXTENDEDNEXTHOPCAPABILITYTUPLE._serialized_end=262
  _EXTENDEDNEXTHOPCAPABILITY._serialized_start=264
  _EXTENDEDNEXTHOPCAPABILITY._serialized_end=346
  _GRACEFULRESTARTCAPABILITYTUPLE._serialized_start=348
  _GRACEFULRESTARTCAPABILITYTUPLE._serialized_end=426
  _GRACEFULRESTARTCAPABILITY._serialized_start=428
  _GRACEFULRESTARTCAPABILITY._serialized_end=539
  _FOUROCTETASNCAPABILITY._serialized_start=541
  _FOUROCTETASNCAPABILITY._serialized_end=578
  _ADDPATHCAPABILITYTUPLE._serialized_start=581
  _ADDPATHCAPABILITYTUPLE._serialized_end=737
  _ADDPATHCAPABILITYTUPLE_MODE._serialized_start=688
  _ADDPATHCAPABILITYTUPLE_MODE._serialized_end=737
  _ADDPATHCAPABILITY._serialized_start=739
  _ADDPATHCAPABILITY._serialized_end=805
  _ENHANCEDROUTEREFRESHCAPABILITY._serialized_start=807
  _ENHANCEDROUTEREFRESHCAPABILITY._serialized_end=839
  _LONGLIVEDGRACEFULRESTARTCAPABILITYTUPLE._serialized_start=841
  _LONGLIVEDGRACEFULRESTARTCAPABILITYTUPLE._serialized_end=942
  _LONGLIVEDGRACEFULRESTARTCAPABILITY._serialized_start=944
  _LONGLIVEDGRACEFULRESTARTCAPABILITY._serialized_end=1044
  _ROUTEREFRESHCISCOCAPABILITY._serialized_start=1046
  _ROUTEREFRESHCISCOCAPABILITY._serialized_end=1075
  _FQDNCAPABILITY._serialized_start=1077
  _FQDNCAPABILITY._serialized_end=1133
  _SOFTWAREVERSIONCAPABILITY._serialized_start=1135
  _SOFTWAREVERSIONCAPABILITY._serialized_end=1188
  _UNKNOWNCAPABILITY._serialized_start=1190
  _UNKNOWNCAPABILITY._serialized_end=1238
# @@protoc_insertion_point(module_scope)



# -*-coding: utf-8 -*-
# vim: sw=4 ts=4 expandtab ai
#
#    Copyright 2022 VK Cloud.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import netaddr


# Bit that indicate that a VLAN id is set.
OFPVID_PRESENT = 0x1000
VLAN_OF_HEADER = "0x8100"


def vid_present(vlan_id):
    """Return VLAN VID with VID_PRESENT flag set"""
    return vlan_id | OFPVID_PRESENT


def push_vlan_action():
    return "push_vlan:%s" % VLAN_OF_HEADER


def get_arp_proxy_responder_action(mac):
    return (
        # Place the source MAC address of the request (The requesting VM)
        # as the new reply's destination MAC address
        "move:NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[],"
        # Put the requested MAC address of the remote VM as source MAC address
        "set_field:%(mac)s->eth_src,"
        # Put an 0x2 code as the type of the ARP message. 0x2 - ARP response.
        "load:0x2->NXM_OF_ARP_OP[],"
        # Place the ARP request's source hardware address (MAC) as this
        # new message"s ARP target / destination hardware address
        "move:NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[],"
        # load arp target protocol address (TPA) of the current packet
        # into register 0 (a special field that is an Open vSwitch
        # extension to OpenFlow)
        "move:NXM_OF_ARP_TPA[]->NXM_NX_REG0[],"
        # Place the ARP request's source protocol / IP address as
        # the new message's ARP destination IP address
        "move:NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[],"
        # load data from register 0 and set it as sender protocol address (SPA)
        "move:NXM_NX_REG0[]->NXM_OF_ARP_SPA[],"
        # Place the requested VM's MAC address as the source MAC address
        "load:%(mac)#x->NXM_NX_ARP_SHA[],"
        # Forward the message back to the port it came in on
        "IN_PORT"
        % {
            "mac": netaddr.EUI(mac, dialect=netaddr.mac_unix_expanded),
        }
    )



# -*-coding: utf-8 -*-
# vim: sw=4 ts=4 expandtab ai
#
#    Copyright 2022 VK Cloud.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from evpn_connector.common import constants
from evpn_connector.ovs import shell


OPENFLOW_PROTO_VERSIONS = ["OpenFlow13"]
VXLAN_PORT_NAME = "vxlan_out"
TUNNEL_TYPE = "vxlan"
DEFAULT_VXLAN_UDP_PORT = 4789

LOG = logging.getLogger(__name__)


class OvSClient(object):
    def __init__(
        self,
        sw_name,
        tmp_flow_file_path,
        vxlan_ofport=None,
        enable_sudo=False,
        ovsvsctl_bin=constants.OVSVSCTL_BIN,
        ovsofctl_bin=constants.OVSOFCTL_BIN,
    ):
        self.vxlan_ofport = vxlan_ofport or constants.VXLAN_PORT_OFPORT
        self.sw_name = sw_name
        self.enable_sudo = enable_sudo
        self.tmp_flow_file_path = tmp_flow_file_path
        self._vsctl_bin = ovsvsctl_bin
        self._ofctl_bin = ovsofctl_bin

    def create_bridge(self):
        cmd = [
            self._vsctl_bin,
            "--may-exist",
            "add-br",
            self.sw_name,
            "--",
            "set",
            "Bridge",
            self.sw_name,
            "protocols={}".format(",".join(OPENFLOW_PROTO_VERSIONS)),
            "fail-mode=secure",
        ]
        return shell.runsh(command=cmd, enable_sudo=self.enable_sudo)

    def create_tun_port(
        self,
        vxlan_source_ip,
        vxlan_udp_port=DEFAULT_VXLAN_UDP_PORT,
        port_name=VXLAN_PORT_NAME,
    ):
        cmd = [
            self._vsctl_bin,
            "--may-exist",
            "add-port",
            self.sw_name,
            port_name,
            "--",
            "set",
            "Interface",
            port_name,
            "type={}".format(TUNNEL_TYPE),
            "options:in_key=flow",
            "options:out_key=flow",
            "options:remote_ip=flow",
            'options:local_ip="{}"'.format(vxlan_source_ip),
            "options:dst_port={}".format(vxlan_udp_port),
            "ofport_request={}".format(self.vxlan_ofport),
        ]
        return shell.runsh(command=cmd, enable_sudo=self.enable_sudo)

    def sync_flows(self, flows):
        # Write data to tmp file
        data = ""
        for flow in sorted(flows):
            data += flow.to_string() + "\n"
        with open(self.tmp_flow_file_path, "w") as group_fl:
            group_fl.write(data)
        LOG.debug(
            "Successfull write %d bytes to tmp flow file %s",
            len(data),
            self.tmp_flow_file_path,
        )
        cmd = [
            self._ofctl_bin,
            "-O",
            OPENFLOW_PROTO_VERSIONS[0],
            "--bundle",
            "replace-flows",
            self.sw_name,
            self.tmp_flow_file_path,
        ]
        return shell.runsh(command=cmd, enable_sudo=self.enable_sudo)



#  coding=utf-8
#
#  Copyright 2025 VK Cloud.
#
#  All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License"); you may
#     not use this file except in compliance with the License. You may obtain
#     a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#     WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#     License for the specific language governing permissions and limitations
#     under the License.

import logging
import sys

from oslo_config import cfg

from evpn_connector.common import constants
from evpn_connector import version

try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

SERVICE_UNDEFINED = "undefined"
SENTRY_TAG_SERVICE = "service"


if sentry_sdk:

    def setup_sentry(service=None):
        if CONF[constants.CONF_GROUP_SENTRY].enabled:
            sentry_sdk.init(
                dsn=CONF[constants.CONF_GROUP_SENTRY].dsn,
                traces_sample_rate=CONF[
                    constants.CONF_GROUP_SENTRY
                ].traces_sample_rate,
                environment=CONF[constants.CONF_GROUP_SENTRY].env,
                attach_stacktrace=True,
                release=version.version_info.release_string(),
            )
            service = service or SERVICE_UNDEFINED
            sentry_sdk.set_tag(SENTRY_TAG_SERVICE, service)

    def capture_exception(error=None, scope=None, **scope_args):
        return sentry_sdk.capture_exception(
            error=error, scope=scope, **scope_args
        )

else:

    def setup_sentry(service=None):
        if CONF[constants.CONF_GROUP_SENTRY].enabled:
            LOG.error("The sentry_sdk library is not installed!")
            sys.exit(5)

    def capture_exception(error=None, scope=None, **scope_args):
        pass



#  coding=utf-8
#
#  Copyright 2025 VK Cloud.
#
#  All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License"); you may
#     not use this file except in compliance with the License. You may obtain
#     a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#     WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#     License for the specific language governing permissions and limitations
#     under the License.

import logging

from loopster.services import softirq

LOG = logging.getLogger(__name__)


class SoftIrqServiceMetrics(softirq.SoftIrqService):
    def _send_event(self, name, event_data, skip=False):
        if self._sender is None:
            self._l(LOG).debug("No sender - skipping event: %s", event_data)
        else:
            if skip:
                self._l(LOG).debug("Skipping send metric: %s", name)
                return
            self._sender.send_duration(name.replace(".", "_"), event_data)

    def _send_step_event(self, event_data):
        self._send_event(
            name="soft_irq.duration.step",
            event_data=event_data["duration"].total_seconds(),
            skip=event_data["skipped"],
        )

    def _send_exc_step_event(self, exc_event_data):
        self._l(LOG).debug("Exception even data: %s", exc_event_data)
        self._sender.send_counter("soft_irq_duration_step.error", value=1)

    def _send_wd_error_event(self, wderr_event_data):
        if wderr_event_data.get("minor"):
            self._l(LOG).debug(
                "Can't use old metrics send: %s", wderr_event_data
            )
        else:
            self._l(LOG).warning(
                "Can't use old metrics send: %s", wderr_event_data
            )



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
#    Copyright 2022 VK Cloud.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import obsender.constants

from oslo_config import cfg

from evpn_connector.common import constants


CONF = cfg.CONF

LOGGING_DOMAIN = "logging"
VICTORIA_METRICS_DOMAIN = obsender.constants.VICTORIA_METRICS_DOMAIN
DAEMON_DOMAIN = "daemon"
GOBGP_DOMAIN = "gobgp"
OVS_DOMAIN = "ovs"
SENTRY_DOMAIN = "sentry"
ANYCAST_DOMAIN = "anycast"

logging_opts = [
    cfg.StrOpt(
        name="config",
        default=constants.DEFAULT_LOGGING_CONFIG_NAME,
        help="Logging subsystem configuration YAML file",
    ),
]

victoria_metrics_opts = [
    cfg.BoolOpt("enabled", default=False, help="Enable sending of metrics"),
    cfg.StrOpt("host", help="Specifies host to send metrics for"),
    cfg.IntOpt("port", help="Specifies port to send metrics for"),
    cfg.StrOpt("prefix", help="Specifies metric prefix for send"),
    cfg.StrOpt("obsender_host", help="Specifies host to send metrics for"),
]

daemon_opts = [
    cfg.IntOpt(
        name="sync_period", default=3, help="Sync bgp announces period"
    ),
    cfg.StrOpt(
        name="configs_dir",
        required=True,
        help="Path to evpn client edge configs",
    ),
]

gobgp_opts = [
    cfg.StrOpt(
        name="gobgp_channel",
        default="localhost:50051",
        help="gobgp grpc host:port",
    ),
    cfg.StrOpt(
        name="source_ip",
        required=True,
        help="Source ip for all EVPN traffic and next_hop for BGP announces",
    ),
    cfg.IntOpt(name="as_number", default=1, help="AS number"),
    cfg.IntOpt(
        name="grpc_timeout_sec",
        default=30,
        help="GoBGP grpc timeout in seconds",
    ),
    cfg.BoolOpt(
        name="policy_enabled",
        default=True,
        help="Enable policy to filter import routes",
    ),
    cfg.StrOpt(
        name="router_mac_type5",
        default=constants.TYPE_5_DEFAULT_ROUTER_MAC_EXTENDED,
        help="Value of RouterMacExtended Ext Communities Attr for Type5",
    ),
]

ovs_opts = [
    cfg.StrOpt(
        name="switch_name", required=True, help="OpenvSwitch switch name"
    ),
    cfg.StrOpt(
        name="tmp_flow_file_path",
        default="/tmp/evpn_tmp_flow_file",
        help="Path to tmp file with ovs flows",
    ),
    cfg.StrOpt(
        name="ovs_vsctl_bin_path",
        default=constants.OVSVSCTL_BIN,
        help="Path to ovs-vsctl binary",
    ),
    cfg.StrOpt(
        name="ovs_ofctl_bin_path",
        default=constants.OVSOFCTL_BIN,
        help="Path to ovs-ofctl binary",
    ),
    cfg.BoolOpt(
        name="enable_sudo",
        default=False,
        help="Enable sudo for all commands to OvS",
    ),
    cfg.IntOpt(
        name="vxlan_udp_port",
        default=4789,
        help="Port number for VXLAN traffic",
    ),
]
anycast_opts = [
    cfg.StrOpt(
        name="anycast_status_file",
        default="/tmp/anycast_status_file",
        help="Path to file with anycast ip statuses from anycast-checker",
    ),
    cfg.IntOpt(
        name="anycast_check_ofport",
        default=constants.ANYCAST_PORT_OFPORT,
        help="OvS ofport number for traffic from anycast-checker",
    ),
    cfg.StrOpt(
        name="anycast_check_mac",
        default=constants.ANYCAST_CHECKS_DST_MAC,
        help="Dst MAC address for anycast-checker traffic",
    ),
]

sentry_opts = [
    cfg.BoolOpt(name="enabled", default=False, help="Set True to enable"),
    cfg.StrOpt(
        name="dsn",
        default="http://<token>@192.168.220.220:9000/<project_id>",
        help="DSN URI to sentry server",
    ),
    cfg.FloatOpt(
        name="traces-sample-rate",
        default=1.0,
        help=(
            "Set traces_sample_rate to 1.0 to capture 100% of "
            "transactions for performance monitoring. We recommend "
            "adjusting this value in production."
        ),
    ),
    cfg.StrOpt(
        name="env",
        default="devenv",
        help=(
            "Environment mark for sentry tagging "
            "and applying env-specific rules"
        ),
    ),
]


def register_logging_opts():
    CONF.register_cli_opts(logging_opts, LOGGING_DOMAIN)


def register_victoria_metrics_opts():
    CONF.register_cli_opts(victoria_metrics_opts, VICTORIA_METRICS_DOMAIN)


def register_daemon_opts():
    CONF.register_cli_opts(daemon_opts, DAEMON_DOMAIN)


def register_gobgp_opts():
    CONF.register_cli_opts(gobgp_opts, GOBGP_DOMAIN)


def register_ovs_opts():
    CONF.register_cli_opts(ovs_opts, OVS_DOMAIN)


def register_sentry_opts(conf=CONF):
    conf.register_opts(sentry_opts, constants.CONF_GROUP_SENTRY)


def register_anycast_opts():
    CONF.register_cli_opts(anycast_opts, ANYCAST_DOMAIN)


def list_opts():
    return [
        (LOGGING_DOMAIN, logging_opts),
        (VICTORIA_METRICS_DOMAIN, victoria_metrics_opts),
        (DAEMON_DOMAIN, daemon_opts),
        (GOBGP_DOMAIN, gobgp_opts),
        (OVS_DOMAIN, ovs_opts),
        (SENTRY_DOMAIN, sentry_opts),
        (ANYCAST_DOMAIN, anycast_opts),
    ]
